#ifndef BmlCaen767VmeDevice_HH
#define BmlCaen767VmeDevice_HH

#include <iostream>
#include <vector>

// hal
#include "VMEDevice.hh"
#include "VMEAddressTable.hh"
#include "VMEBusAdapterInterface.hh"

// dual/inc/bml
#include "BmlCaen767RunData.hh"
#include "BmlCaen767ConfigurationData.hh"
#include "BmlCaen767EventData.hh"


class BmlCaen767VmeDevice : public VMEDevice {
public:
  BmlCaen767VmeDevice(VMEAddressTable &t, VMEBusAdapterInterface &b, unsigned short a=0x00c5)
    : VMEDevice(t,b,a<<16) {
  }

  bool alive() {
    long unsigned value;
    try {
      UtlPack bid(0);
      for(unsigned i(0);i<7;i++) {
	read("BoardId",&value,2*i);
	std::cout << "BoardId+" << std::setw(2) << i
		  << " = " << printHex((unsigned)value) << std::endl;
	//bid.byte(i,value);
      }
      return UtlPack(0x000002ff)==bid;

    } catch ( HardwareAccessException& e ) {
      cout << "*** HA Exception occurred : " << e.what() << endl;
      return false;

    } catch ( std::exception e ) {
      cout << "*** Unknown exception occurred" << endl;
      return false;
    }
  }

  bool reset() {
    write("SingleShotReset",0);    
    sleep(2);
    // Set any constants here
    return true;
  }

  bool softTrigger() {
    write("SoftwareTrigger",0);
    return true;
  }

  bool clear() {
    write("Clear",0);
    return true;
  }

  bool readRunData(BmlCaen767RunData &d) {
    long unsigned value;

    UtlPack ba;
    read("AddressDecoder24",&value);
    ba.byte(0,value&0xff);
    read("AddressDecoder32",&value);
    ba.byte(1,value&0xff);
    d.baseAddress(ba.halfWord(0));

    std::string mid(10);
    for(unsigned i(0);i<5;i++) {
      read("ManufacturerId",&value,2*i);
      mid[2*i  ]= value    &0xff;
      mid[2*i+1]=(value>>8)&0xff;
    }
    d.manufacturerId(mid);
    
    std::string bid(14);
    for(unsigned i(0);i<7;i++) {
      read("BoardId",&value,2*i);
      bid[2*i  ]= value    &0xff;
      bid[2*i+1]=(value>>8)&0xff;
    }
    d.boardId(mid);

    read("RevisionId",&value);
    d.revisionId(value&0xffff);

    read("SlaveParameters",&value);
    std::cout << "SlaveParameters " << printHex(value) << std::endl;
    //d.slaveParameters(value&0xffff);

    read("InteruptCapability",&value);
    std::cout << "InteruptCapability " << printHex(value) << std::endl;
    //d.interuptCapability(value&0xffff);

    read("F0AccessWidth",&value);
    std::cout << "F0AccessWidth " << printHex(value) << std::endl;
    //d.f0AccessWidth(value&0xffff);

    read("F1AccessWidth",&value);
    std::cout << "F1AccessWidth " << printHex(value) << std::endl;
    //d.f1AccessWidth(value&0xffff);

    std::string f0c(30);
    for(unsigned i(0);i<15;i++) {
      read("F0AmCodeMask",&value,2*i);
      f0c[2*i  ]= value    &0xff;
      f0c[2*i+1]=(value>>8)&0xff;
    }
    std::cout << "F0AmCodeMask " << f0c << std::endl;
    //d.f0AmCodeMask(f0c);

    std::string f1c(30);
    for(unsigned i(0);i<15;i++) {
      read("F1AmCodeMask",&value,2*i);
      f1c[2*i  ]= value    &0xff;
      f1c[2*i+1]=(value>>8)&0xff;
    }
    std::cout << "F1AmCodeMask " << f1c << std::endl;
    //d.f1AmCodeMask(f1c);

    std::string f0a(14);
    for(unsigned i(0);i<7;i++) {
      read("F0AddressMask",&value,2*i);
      f0a[2*i  ]= value    &0xff;
      f0a[2*i+1]=(value>>8)&0xff;
    }
    std::cout << "F0AddressMask " << f0a << std::endl;
    //d.f0AddressMask(f0a);

    std::string f1a(14);
    for(unsigned i(0);i<7;i++) {
      read("F0AddressMask",&value,2*i);
      f1a[2*i  ]= value    &0xff;
      f1a[2*i+1]=(value>>8)&0xff;
    }
    std::cout << "F1AddressMask " << f1a << std::endl;
    //d.f1AddressMask(f1a);

    read("SerialNumber",&value);
    d.serialNumber(value&0xffff);

    return true;
  }

  bool writeConfigurationData(const BmlCaen767ConfigurationData &d) {
    write("CSR",d.csr());
    write("CSR",d.csr()|0x00800000);
    return true;
  }

  bool readConfigurationData(BmlCaen767ConfigurationData &d) {
    long unsigned value;
    read("CSR",&value);
    d.csr(value);
    return true;
  }

  bool readEventData(BmlCaen767EventData &d, unsigned n=0) {
    long unsigned value;
    read("CSR",&value);
    d.csr(value);

    d.numberOfWords(0);
    unsigned nEvents((value>>24)&0x1f);
    //cout << "Number of events = " << nEvents << endl;
    if(n>=nEvents) return true;

    unsigned nWords(256);
    unsigned *p(d.data());
    for(unsigned i(0);i<4 && nWords==256;i++) {
      readBlock("DataBlt",256,(char*)(p+64*i),HAL_DO_INCREMENT,0x0400*n+256*i);
      for(unsigned j(0);j<64 && nWords==256;j++) {
	//cout << "Word " << 64*i+j << printHex(p[64*i+j]) << endl;

	if((p[64*i+j]&0x00800000)==0) {
	  nWords=64*i+j;
	  if((p[64*i+j]&0x00200000)==0) nWords++;
	}
      }
    }

    d.numberOfWords(nWords);

    return true;
  }

  void print(std::ostream &o) {
    o << "BmlCaen767VmeDevice::print()" << std::endl;
    /*
    long unsigned value;
    read("FirmwareId",&value);
    o << hex << "  FirmwareId = 0x" << value << dec << std::endl;
    */
    o << std::endl;

  }

private:
  bool writeOpcode(unsigned char c, unsigned char o) {
    return writeOpcode((c<<8)+o);
  }

  bool writeOpcode(unsigned short d) {
    long unsigned value;
    for(unsigned i(0);i<1000000;i++) {
      read("OpcodeHandshake",&value);
      std::cout << "writeOpcode " << printHex(value) << std::endl;

      // Check for write bit
      if((value&2)!=0) {
	usleep(10000);
	write("OpcodeRegister",d);
	return true;
      }
    }
    return false;
  }

  bool readOpcode(unsigned short &d) {
    long unsigned value;
    for(unsigned i(0);i<1000000;i++) {
      read("OpcodeHandshake",&value);
      std::cout << "readOpcode " << printHex(value) << std::endl;

      // Check for read bit
      if((value&1)!=0) {
	usleep(10000);
	read("OpcodeRegister",&value);
	d=value&0xffff;
	return true;
      }
    }
    return false;
  }
};

#endif
